/*
  "AD9850_51"
  library providing functions to control DDS modules based on AD9850 or AD9851 in serial mode (data pin == D7)
  (AD9850 is usually clocked by a 125MHz crystal whereas AD9851 uses a 30MHz crystal and internal 6x clock
   multiplier)
  V1.0 / Copyright (C) 2019, T.Rode (DL1DUZ)

  Permission is granted to use, copy, modify, and distribute this software
  and documentation for non-commercial purposes.

  Acknowledgements:
  Most algorithms were derived from Christophe Caiveaus library "Christophe Caiveau f4goj@free.fr" 
*/

//################################################################################################################
//includes
//################################################################################################################

#include <AD9850_51.h>

//################################################################################################################
//declarations
//################################################################################################################


//################################################################################################################
//constructor
//################################################################################################################

// W_CLK_PIN, FQ_UD_PIN, DATA_PIN, RESET_PIN, CLOCK_FREQUENCY[Hz], Flag, indicating that chip is AD9851
AD9850_51::AD9850_51(unsigned char w_clk_pin, unsigned char fq_ud_pin, unsigned char data_pin,
               unsigned char reset_pin, unsigned long clk_frequency, unsigned char is_AD9851) {
// assign and initialize data pins
  W_CLK_PIN = w_clk_pin;
  FQ_UD_PIN = fq_ud_pin;
  DATA_PIN = data_pin;
  RESET_PIN = reset_pin;

  pinMode(W_CLK_PIN, OUTPUT);
  pinMode(FQ_UD_PIN, OUTPUT);
  pinMode(DATA_PIN, OUTPUT);
  pinMode(RESET_PIN, OUTPUT);

  is_9851 = is_AD9851;

// turn on serial comm. mode
  ddsPulse(RESET_PIN);
  ddsPulse(W_CLK_PIN);
  ddsPulse(FQ_UD_PIN);

// turn off the DDS
  setPhase(0, 0, 1);

// set the system clock frequency
  FC = clk_frequency;
}

//################################################################################################################
//functions
//################################################################################################################

// support function that calculates the DDS clock frequency at nominal temperatur by comparing a set output to the
// actually measured one; e.g. if a 180MHz-nominal DDS is set to 10MHz and the measured output is 9.9997MHz
// "calculateClockFrequency(10000000, 9999700, 180000000)" returns 179994600Hz = 179.9946MHz; for good
// numeric accuracy the test frequency should be chosen as high as possible
unsigned long AD9850_51::calculateClockFrequency(unsigned long f_set, unsigned long f_real, unsigned long clock) {
  uint64_t dummy = f_real*clock;
  dummy /= f_set;

  return (unsigned long) dummy;
}

//################################################################################################################

// calculates the phase-value for a given frequency and DDS clock offset [Hz] caused by offset in crystal temperature
unsigned long AD9850_51::calculatePhaseValue(unsigned long frequency, int clockOffset) {
  uint64_t dummy = frequency;
  dummy <<= 32;
  dummy /= (FC + clockOffset);

  return (unsigned long) dummy; 
}

//################################################################################################################

// calculates the frequency [Hz] for a give phase-value and clock offset [Hz] caused by offset in crystal temperature
unsigned long AD9850_51::calculateFrequency(unsigned long phase, int clockOffset) {
  uint64_t dummy = phase;
  dummy *= (FC + clockOffset);
  dummy >>= 32;

  return (unsigned long) dummy; 
}

//################################################################################################################

inline void AD9850_51::ddsPulse(uint8_t pin) {
  digitalWrite(pin, 1);
  digitalWrite(pin, 0);
}

//################################################################################################################

inline void AD9850_51::ddsShiftOut(uint8_t data) {
  shiftOut(DATA_PIN, W_CLK_PIN, LSBFIRST, data);
}

//################################################################################################################

// sets deltaphase (phase-stepwidth) and initial phase; initial phase coded on 5 bits (value ranging from 0...32,
// each step representing an angle of 11.25 degree)
// if shutdown is true, DDS will shut down, reducing power consumption and dissipated heat to a minimum
void AD9850_51::setPhase(unsigned long deltaphase, unsigned char phase, boolean shutdown) {
  for (uint8_t i=0; i<4; i++, (deltaphase>>=8)) {
   ddsShiftOut(deltaphase & 0xFF);
  }

  phase <<= 3; // shift phase-word into right position
  if(is_9851) phase += 1; // set 6x clock multiplier flag in case chip is AD9851
  if(shutdown) {
    if(!dds_is_down) {
      dds_is_down = 1;
      phase += 4; // set shutdown flag
    }
  }
  else { dds_is_down = 0; }

  ddsShiftOut(phase);
  ddsPulse(FQ_UD_PIN);
}
